/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: HTS_engine.h,v 1.14 2008/03/07 07:06:15 uratec Exp $ */

#include <stdio.h>

/*  -------------------------- common -----------------------------  */

typedef int HTS_Boolean;
#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif                          /* !TRUE */

typedef int HTS_VocoderType;
#define HTS_VOCODER_TYPE_MCP      0
#define HTS_VOCODER_TYPE_LSP      1

/*  -------------------------- model ------------------------------  */

/* HTS_DWin: structure for regression window */
typedef struct _HTS_Window {
   int size;                    /* number of static + deltas */
   int *l_width;                /* left width */
   int *r_width;                /* right width */
   double **coefficient;        /* coefficient */
   int max_width;               /* max width */
} HTS_Window;

/* HTS_Pattern: structure for pattern */
typedef struct _HTS_Pattern {
   char *string;                /* pattern */
   struct _HTS_Pattern *next;   /* link to next pattern */
} HTS_Pattern;

/* HTS_Question: structure for questions */
typedef struct _HTS_Question {
   char *string;                /* name of this question */
   HTS_Pattern *head;           /* link to head of pattern list */
   struct _HTS_Question *next;  /* link to next question */
} HTS_Question;

/* HTS_Node: structure for node of decision tree */
typedef struct _HTS_Node {
   int index;                   /* index of this node */
   int pdf;                     /* index of pdf for this node  ( leaf node only ) */
   struct _HTS_Node *yes;       /* link to child node (yes) */
   struct _HTS_Node *no;        /* link to child node (no) */
   struct _HTS_Node *next;      /* link to next node */
   HTS_Question *quest;         /* question applied at this node */
} HTS_Node;

/* HTS_Tree: structure for each decision tree */
typedef struct _HTS_Tree {
   HTS_Pattern *head;           /* link to head of pattern list for this tree */
   struct _HTS_Tree *next;      /* link to next tree */
   HTS_Node *root;              /* root node of this decision tree */
   int state;                   /* state of this tree */
} HTS_Tree;

/* HTS_Model: structure for model */
typedef struct _HTS_Model {
   int vector_length;           /* vector length */
   int ntree;                   /* number of tree */
   int *npdf;                   /* number of pdf per tree */
   double ***pdf;               /* pdfs */
   HTS_Tree *tree;              /* decision trees */
   HTS_Question *question;      /* questions */
} HTS_Model;

/* HTS_Stream: structure for stream */
typedef struct _HTS_Stream {
   int vector_length;           /* vector_length */
   HTS_Model *model;            /* models */
   HTS_Window window;           /* dynamic window */
   HTS_Boolean msd_flag;        /* MSD flag */
   int interpolation_size;      /* size for interpolation */
} HTS_Stream;

/* HTS_ModelSet: structure for HMM set */
typedef struct _HTS_ModelSet {
   HTS_Stream duration;         /* duration pdfs and trees */
   HTS_Stream *stream;          /* parameter pdfs, trees and windows */
   HTS_Stream *gv;              /* GV pdpfs */
   int nstate;                  /* number of state */
   int nstream;                 /* number of stream */
} HTS_ModelSet;

/*  ----------------------- model method --------------------------  */

/* HTS_ModelSet_initialize: initialize model set */
void HTS_ModelSet_initialize(HTS_ModelSet * ms, int nstream);

/* HTS_ModelSet_load_duration: load duration model and number of state */
void HTS_ModelSet_load_duration(HTS_ModelSet * ms, FILE ** pdf_fp,
                                FILE ** tree_fp, int interpolation_size);

/* HTS_ModelSet_load_parameter: load parameter model */
void HTS_ModelSet_load_parameter(HTS_ModelSet * ms, FILE ** pdf_fp,
                                 FILE ** tree_fp, FILE ** win_fp,
                                 int stream_index, HTS_Boolean msd_flag,
                                 int window_size, int interpolation_size);

/* HTS_ModelSet_load_gv: load GV model */
void HTS_ModelSet_load_gv(HTS_ModelSet * ms, FILE ** pdf_fp, int stream_index,
                          int interpolation_size);

/* HTS_ModelSet_get_nstate: get number of state */
inline int HTS_ModelSet_get_nstate(HTS_ModelSet * ms);

/* HTS_ModelSet_get_nstream: get number of stream */
inline int HTS_ModelSet_get_nstream(HTS_ModelSet * ms);

/* HTS_ModelSet_get_vector_length: get vector length */
inline int HTS_ModelSet_get_vector_length(HTS_ModelSet * ms, int stream_index);

/* HTS_ModelSet_is_msd: get MSD flag */
inline HTS_Boolean HTS_ModelSet_is_msd(HTS_ModelSet * ms, int stream_index);

/* HTS_ModelSet_get_window_size: get dynamic window size */
inline int HTS_ModelSet_get_window_size(HTS_ModelSet * ms, int stream_index);

/* HTS_ModelSet_get_window_left_width: get left width of dynamic window */
inline int HTS_ModelSet_get_window_left_width(HTS_ModelSet * ms,
                                              int stream_index,
                                              int window_index);

/* HTS_ModelSet_get_window_right_width: get right width of dynamic window */
inline int HTS_ModelSet_get_window_right_width(HTS_ModelSet * ms,
                                               int stream_index,
                                               int window_index);

/* HTS_ModelSet_get_window_coefficient: get coefficient of dynamic window */
inline double HTS_ModelSet_get_window_coefficient(HTS_ModelSet * ms,
                                                  int stream_index,
                                                  int window_index,
                                                  int coefficient_index);

/* HTS_ModelSet_get_window_max_width: get max width of dynamic window */
inline int HTS_ModelSet_get_window_max_width(HTS_ModelSet * ms,
                                             int stream_index);

/* HTS_ModelSet_get_duration_interpolation_size: get interpolation size (duration model) */
inline int HTS_ModelSet_get_duration_interpolation_size(HTS_ModelSet * ms);

/* HTS_ModelSet_get_parameter_interpolation_size: get interpolation size (parameter model) */
inline int HTS_ModelSet_get_parameter_interpolation_size(HTS_ModelSet * ms,
                                                         int stream_index);

/* HTS_ModelSet_get_gv_interpolation_size: get interpolation size (GV model) */
inline int HTS_ModelSet_get_gv_interpolation_size(HTS_ModelSet * ms,
                                                  int stream_index);

/* HTS_ModelSet_use_gv: get GV flag */
inline HTS_Boolean HTS_ModelSet_use_gv(HTS_ModelSet * ms, int index);

/* HTS_ModelSet_get_duration: get duration using interpolation weight */
void HTS_ModelSet_get_duration(HTS_ModelSet * ms, char *string, double *mean,
                               double *vari, double *iw);

/* HTS_ModelSet_get_parameter: get parameter using interpolation weight */
void HTS_ModelSet_get_parameter(HTS_ModelSet * ms, char *string, double *mean,
                                double *vari, double *msd, int stream_index,
                                int state_index, double *iw);

/* HTS_ModelSet_get_gv: get GV using interpolation weight */
void HTS_ModelSet_get_gv(HTS_ModelSet * ms, double *mean, double *vari,
                         int stream_index, double *iw);

/* HTS_ModelSet_clear: free model set */
void HTS_ModelSet_clear(HTS_ModelSet * ms);

/*  -------------------------- label ------------------------------  */

/* HTS_LabelString: structure for individual label string with time infomation */
typedef struct _HTS_LabelString {
   struct _HTS_LabelString *next;       /* link to next label string */
   char *name;                  /* label string */
   HTS_Boolean frame_flag;      /* specified frame length flag */
   int frame;                   /* specified frame length */
   HTS_Boolean speed_flag;      /* specified speech speed flag */
   double speed;                /* specified speech speed */
} HTS_LabelString;

/* HTS_Label: structure for label */
typedef struct _HTS_Label {
   HTS_LabelString *head;       /* link to head of label string */
   int size;                    /* number of label string */
} HTS_Label;

/*  ----------------------- label method --------------------------  */

/* HTS_Label_initialize: initialize label */
void HTS_Label_initialize(HTS_Label * label);

/* HTS_Label_load_from_fn: load label from file name */
void HTS_Label_load_from_fn(HTS_Label * label, int sampling_rate, int fperiod,
                            char *fn);

/* HTS_Label_load_from_fp: load label list from file pointer */
void HTS_Label_load_from_fp(HTS_Label * label, int sampling_rate, int fperiod,
                            FILE * fp);

/* HTS_Label_load_from_string: load label from string */
void HTS_Label_load_from_string(HTS_Label * label, int sampling_rate,
                                int fperiod, char *data);

/* HTS_Label_load_from_string_list: load label list from string list */
void HTS_Label_load_from_string_list(HTS_Label * label, int sampling_rate,
                                     int fperiod, char **data, int size);

/* HTS_Label_set_speech_speed: set speech speed */
void HTS_Label_set_speech_speed(HTS_Label * label, int string_index, double f);

/* HTS_Label_set_frame: set frame length */
void HTS_Label_set_frame(HTS_Label * label, int string_index, int i);

/* HTS_Label_get_size: get number of label string */
inline int HTS_Label_get_size(HTS_Label * label);

/* HTS_Label_get_string: get label string */
char *HTS_Label_get_string(HTS_Label * label, int string_index);

/* HTS_Label_get_frame_specified_flag: get frame specified flag */
HTS_Boolean HTS_Label_get_frame_specified_flag(HTS_Label * label,
                                               int string_index);

/* HTS_Label_get_frame: get frame length */
int HTS_Label_get_frame(HTS_Label * label, int string_index);

/* HTS_Label_get_speed_specified_flag: get speech speed specified flag */
HTS_Boolean HTS_Label_get_speed_specified_flag(HTS_Label * label,
                                               int string_index);

/* HTS_Label_get_speed: get speech speed */
double HTS_Label_get_speed(HTS_Label * label, int string_index);

/* HTS_Label_clear: free label */
void HTS_Label_clear(HTS_Label * label);

/*  -------------------------- sstream ----------------------------  */

/* HTS_SStream: structure for individual state sequence */
typedef struct _HTS_SStream {
   int vector_length;           /* vector length */
   double **mean;               /* mean vector sequence */
   double **vari;               /* variance vector sequence */
   double *msd;                 /* MSD parameter sequence */
   int win_size;                /* number of static + deltas of dynamic window */
   int *win_l_width;            /* left width of dynamic window */
   int *win_r_width;            /* right width of dynamic window */
   double **win_coefficient;    /* coefficient of dynamic window */
   int win_max_width;           /* max width of dynamic window */
   double *gv_mean;             /* mean of GV */
   double *gv_vari;             /* variance of GV */
} HTS_SStream;

/* HTS_SStreamSet: structure for state sequences */
typedef struct _HTS_SStreamSet {
   HTS_SStream *sstream;        /* state stream */
   int nstream;                 /* number of stream */
   int nstate;                  /* number of state */
   int *duration;               /* duration sequence */
   int total_state;             /* total number of state */
   int total_frame;             /* total number of frame */
} HTS_SStreamSet;

/*  ----------------------- sstream method ------------------------  */

/* HTS_SStreamSet_initialize: initialize state stream set */
void HTS_SStreamSet_initialize(HTS_SStreamSet * sss);

/* HTS_SStreamSet_create: parse label and determine state duration */
void HTS_SStreamSet_create(HTS_SStreamSet * sss, HTS_ModelSet * ms,
                           HTS_Label * label, double rho, double *duration_iw,
                           double **parameter_iw, double **gv_iw);

/* HTS_SStreamSet_get_nstream: get number of stream */
inline int HTS_SStreamSet_get_nstream(HTS_SStreamSet * sss);

/* HTS_SStreamSet_get_vector_length: get vector length */
inline int HTS_SStreamSet_get_vector_length(HTS_SStreamSet * sss,
                                            int stream_index);

/* HTS_SStreamSet_is_msd: get MSD flag */
inline HTS_Boolean HTS_SStreamSet_is_msd(HTS_SStreamSet * sss,
                                         int stream_index);

/* HTS_SStreamSet_get_total_state: get total number of state */
inline int HTS_SStreamSet_get_total_state(HTS_SStreamSet * sss);

/* HTS_SStreamSet_get_total_frame: get total number of frame */
inline int HTS_SStreamSet_get_total_frame(HTS_SStreamSet * sss);

/* HTS_SStreamSet_get_msd: get msd parameter */
inline double HTS_SStreamSet_get_msd(HTS_SStreamSet * sss, int stream_index,
                                     int state_index);

/* HTS_SStreamSet_window_size: get dynamic window size */
inline int HTS_SStreamSet_get_window_size(HTS_SStreamSet * sss,
                                          int stream_index);

/* HTS_SStreamSet_get_window_left_width: get left width of dynamic window */
inline int HTS_SStreamSet_get_window_left_width(HTS_SStreamSet * sss,
                                                int stream_index,
                                                int window_index);

/* HTS_SStreamSet_get_window_right_width: get right width of dynamic window */
inline int HTS_SStreamSet_get_window_right_width(HTS_SStreamSet * sss,
                                                 int stream_index,
                                                 int window_index);

/* HTS_SStreamSet_get_window_coefficient: get coefficient of dynamic window */
inline double HTS_SStreamSet_get_window_coefficient(HTS_SStreamSet * sss,
                                                    int stream_index,
                                                    int window_index,
                                                    int coefficient_index);

/* HTS_SStreamSet_get_window_max_width: get max width of dynamic window */
inline int HTS_SStreamSet_get_window_max_width(HTS_SStreamSet * sss,
                                               int stream_index);

/* HTS_SStreamSet_use_gv: get GV flag */
inline HTS_Boolean HTS_SStreamSet_use_gv(HTS_SStreamSet * sss,
                                         int stream_index);

/* HTS_SStreamSet_get_duration: get state duration */
inline int HTS_SStreamSet_get_duration(HTS_SStreamSet * sss, int state_index);

/* HTS_SStreamSet_get_mean: get mean parameter */
inline double HTS_SStreamSet_get_mean(HTS_SStreamSet * sss, int stream_index,
                                      int state_index, int vector_index);

/* HTS_SStreamSet_set_mean: set mean parameter */
inline void HTS_SStreamSet_set_mean(HTS_SStreamSet * sss, int stream_index,
                                    int state_index, int vector_index,
                                    double f);

/* HTS_SStreamSet_get_vari: get variance parameter */
inline double HTS_SStreamSet_get_vari(HTS_SStreamSet * sss, int stream_index,
                                      int state_index, int vector_index);

/* HTS_SStreamSet_set_vari: set variance parameter */
inline void HTS_SStreamSet_set_vari(HTS_SStreamSet * sss, int stream_index,
                                    int state_index, int vector_index,
                                    double f);

/* HTS_SStreamSet_get_gv_mean: get GV mean parameter */
inline double HTS_SStreamSet_get_gv_mean(HTS_SStreamSet * sss, int stream_index,
                                         int vector_index);

/* HTS_SStreamSet_get_gv_mean: get GV variance parameter */
inline double HTS_SStreamSet_get_gv_vari(HTS_SStreamSet * sss, int stream_index,
                                         int vector_index);

/* HTS_SStreamSet_clear: free state stream set */
void HTS_SStreamSet_clear(HTS_SStreamSet * sss);

/*  -------------------------- pstream ----------------------------  */

/* HTS_SMatrices: structure for matrices to generate sequence of speech parameter vectors */
typedef struct _HTS_SMatrices {
   double **mean;               /* sequence of mean vector */
   double **ivar;               /* sequence of invarsed variance vector */
   double *g;                   /* for forward substitution */
   double **wuw;                /* W' U^-1 W  */
   double *wum;                 /* W' U^-1 mu */
} HTS_SMatrices;

/* HTS_PStream: structure for parameter generation */
typedef struct _HTS_PStream {
   int vector_length;           /* observation vector length (include static and dynamic features) */
   int static_length;           /* static features length */
   int length;                  /* stream length */
   int width;                   /* width of dynamic window */
   double **par;                /* output parameter vector */
   HTS_SMatrices sm;            /* matrices for parameter generation */
   int win_size;                /* number of static + deltas of dynamic window */
   int *win_l_width;            /* left width of dynamic window */
   int *win_r_width;            /* right width of dynamic window */
   double **win_coefficient;    /* coefficient of dynamic window */
   HTS_Boolean *msd_flag;       /* TRUE/FALSE sequence for MSD */
   double *gv_buff;             /* GV buffer */
   double *gv_mean;             /* GV mean */
   double *gv_vari;             /* GV variance */
   double gv_weight;            /* GV weight */
} HTS_PStream;

/* HTS_PStreamSet: structure for parameter stream set */
typedef struct _HTS_PStreamSet {
   HTS_PStream *pstream;        /* parameter stream */
   int nstream;                 /* number of stream */
   int total_frame;             /* total frame length */
} HTS_PStreamSet;

/*  ----------------------- pstream method ------------------------  */

/* HTS_PStreamSet_initialize: initialize parameter stream set */
void HTS_PStreamSet_initialize(HTS_PStreamSet * pss);

/* HTS_PStreamSet_create: parameter generation using GV weight */
void HTS_PStreamSet_create(HTS_PStreamSet * pss, HTS_SStreamSet * sss,
                           double *msd_threshold, double *gv_weight);

/* HTS_PStreamSet_get_nstream: get number of stream */
inline int HTS_PStreamSet_get_nstream(HTS_PStreamSet * pss);

/* HTS_PStreamSet_get_static_length: get static features length */
inline int HTS_PStreamSet_get_static_length(HTS_PStreamSet * pss,
                                            int stream_index);

/* HTS_PStreamSet_get_total_frame: get total number of frame */
inline int HTS_PStreamSet_get_total_frame(HTS_PStreamSet * pss);

/* HTS_PStreamSet_get_parameter: get parameter */
inline double HTS_PStreamSet_get_parameter(HTS_PStreamSet * pss,
                                           int stream_index, int frame_index,
                                           int vector_index);

/* HTS_PStreamSet_get_msd_flag: get generated MSD flag per frame */
inline HTS_Boolean HTS_PStreamSet_get_msd_flag(HTS_PStreamSet * pss,
                                               int stream_index,
                                               int frame_index);

/* HTS_PStreamSet_is_msd: get MSD flag */
inline HTS_Boolean HTS_PStreamSet_is_msd(HTS_PStreamSet * pss,
                                         int stream_index);

/* HTS_PStreamSet_clear: free parameter stream set */
void HTS_PStreamSet_clear(HTS_PStreamSet * pss);

/*  -------------------------- gstream ----------------------------  */

/* HTS_GStream: structure for generated parameter */
typedef struct _HTS_GStream {
   int static_length;           /* static features length */
   double **par;                /* generated parameter sequence */
} HTS_GStream;

/* HTS_GStreamSet: generated parameter stream set */
typedef struct _HTS_GStreamSet {
   int total_nsample;           /* total number of sample */
   int total_frame;             /* total frame length */
   int nstream;                 /* number of stream */
   HTS_GStream *gstream;        /* generated parameter stream */
   short *gspeech;              /* generated speech */
} HTS_GStreamSet;

/*  ----------------------- gstream method ------------------------  */

/* HTS_GStreamSet_initialize: initialize generated parameter stream set */
void HTS_GStreamSet_initialize(HTS_GStreamSet * gss);

/* HTS_GStreamSet_create: generate speech */
void HTS_GStreamSet_create(HTS_GStreamSet * gss, HTS_PStreamSet * pss,
                           HTS_VocoderType type, int stage, int sampling_rate,
                           int fperiod, double alpha, double beta,
                           int audio_buff_size);

/* HTS_GStreamSet_get_total_nsample: get total number of sample */
inline int HTS_GStreamSet_get_total_nsample(HTS_GStreamSet * gss);

/* HTS_GStreamSet_get_total_frame: get total number of frame */
inline int HTS_GStreamSet_get_total_frame(HTS_GStreamSet * gss);

/* HTS_GStreamSet_get_static_length: get static features length */
inline int HTS_GStreamSet_get_static_length(HTS_GStreamSet * gss,
                                            int stream_index);

/* HTS_GStreamSet_get_speech: get synthesized speech parameter */
inline short HTS_GStreamSet_get_speech(HTS_GStreamSet * gss, int sample_index);

/* HTS_GStreamSet_get_parameter: get generated parameter */
inline double HTS_GStreamSet_get_parameter(HTS_GStreamSet * gss,
                                           int stream_index, int frame_index,
                                           int vector_index);

/* HTS_GStreamSet_clear: free generated parameter stream set */
void HTS_GStreamSet_clear(HTS_GStreamSet * gss);

/*  -------------------------- engine -----------------------------  */

/* HTS_Global: structure for global parameter */
typedef struct _HTS_Global {
   int stage;                   /* stage==0.0 ? Gamma=0.0 : Gamma=-1.0/stage */
   int sampling_rate;           /* sampling rate */
   int fperiod;                 /* frame period */
   double rho;                  /* control duration parameter */
   double alpha;                /* all-pass constant */
   double beta;                 /* postfiltering coefficient */
   int audio_buff_size;         /* audio buffer size (for audio device) */
   HTS_VocoderType type;        /* vocode type */
   double *msd_threshold;       /* MSD threshold list */
   double *duration_iw;         /* interpolation weights for duration */
   double **parameter_iw;       /* interpolation weights for parameter */
   double **gv_iw;              /* interpolation weights for GV */
   double *gv_weight;           /* GV weights */
} HTS_Global;

typedef struct _HTS_Engine {
   HTS_Global global;           /* global parameter */
   HTS_ModelSet ms;             /* model set */
   HTS_Label label;             /* label */
   HTS_SStreamSet sss;          /* state stream set */
   HTS_PStreamSet pss;          /* parameter stream set */
   HTS_GStreamSet gss;          /* generated parameter stream set */
} HTS_Engine;

/*  ----------------------- engine method -------------------------  */

/* HTS_Engine_initialize: initialize engine */
void HTS_Engine_initialize(HTS_Engine * engine, int nstream);

/* HTS_engine_load_duration_from_fn: load duration pdfs ,trees and number of state from file names */
void HTS_Engine_load_duration_from_fn(HTS_Engine * engine, char **pdf_fn,
                                      char **tree_fn, int interpolation_size);

/* HTS_Engine_load_duration_from_fp: load duration pdfs, trees and number of state from file pointers */
void HTS_Engine_load_duration_from_fp(HTS_Engine * engine, FILE ** pdf_fp,
                                      FILE ** tree_fp, int interpolation_size);

/* HTS_Engine_load_parameter_from_fn: load parameter pdfs, trees and windows from file names */
void HTS_Engine_load_parameter_from_fn(HTS_Engine * engine, char **pdf_fn,
                                       char **tree_fn, char **win_fn,
                                       int stream_index, HTS_Boolean msd_flag,
                                       int window_size, int interpolation_size);

/* HTS_Engine_load_parameter_from_fp: load parameter pdfs, trees and windows from file pointers */
void HTS_Engine_load_parameter_from_fp(HTS_Engine * engine, FILE ** pdf_fp,
                                       FILE ** tree_fp, FILE ** win_fp,
                                       int stream_index, HTS_Boolean msd_flag,
                                       int window_size, int interpolation_size);

/* HTS_Engine_load_gv_from_fn: load GV pdfs from file names */
void HTS_Engine_load_gv_from_fn(HTS_Engine * engine, char **pdf_fn,
                                int stream_index, int interpolation_size);

/* HTS_Engine_load_gv_from_fp: load GV pdfs from file pointers */
void HTS_Engine_load_gv_from_fp(HTS_Engine * engine, FILE ** pdf_fp,
                                int stream_index, int interpolation_size);

/* HTS_Engine_set_sampling_rate: set sampling rate */
void HTS_Engine_set_sampling_rate(HTS_Engine * engine, int i);

/* HTS_Engine_set_fperiod: set frame shift */
void HTS_Engine_set_fperiod(HTS_Engine * engine, int i);

/* HTS_Engine_set_alpha: set alpha */
void HTS_Engine_set_alpha(HTS_Engine * engine, double f);

/* HTS_Engine_set_gamma: set gamma (gamma = -1 / i : if i=0 then gamma=0) */
void HTS_Engine_set_gamma(HTS_Engine * engine, int i);

/* HTS_Engine_set_beta: set beta */
void HTS_Engine_set_beta(HTS_Engine * engine, double f);

/* HTS_Engine_set_rho: set control duration parameter */
void HTS_Engine_set_rho(HTS_Engine * engine, double f);

/* HTS_Engine_set_audio_buff_size: set audio buffer size */
void HTS_Engine_set_audio_buff_size(HTS_Engine * engine, int i);

/* HTS_engine_set_vocoder_type: set vocoder type */
void HTS_Engine_set_vocoder_type(HTS_Engine * engine, HTS_VocoderType type);

/* HTS_Egnine_set_msd_threshold: set MSD threshold */
void HTS_Engine_set_msd_threshold(HTS_Engine * engine, int stream_index,
                                  double f);

/* HTS_Engine_set_duration_interpolation_weight: set interpolation weight for duration */
void HTS_Engine_set_duration_interpolation_weight(HTS_Engine * engine,
                                                  int interpolation_index,
                                                  double f);

/* HTS_Engine_set_parameter_interpolation_weight: set interpolation weight for parameter */
void HTS_Engine_set_parameter_interpolation_weight(HTS_Engine * engine,
                                                   int stream_index,
                                                   int interpolation_index,
                                                   double f);

/* HTS_Engine_set_gv_interpolation_weight: set interpolation weight for GV */
void HTS_Engine_set_gv_interpolation_weight(HTS_Engine * engine,
                                            int stream_index,
                                            int interpolation_index, double f);

/* HTS_Engine_set_gv_weight: set GV weight */
void HTS_Engine_set_gv_weight(HTS_Engine * engine, int stream_index, double f);

/* HTS_Engine_load_label_from_fn: load label from file pointer */
void HTS_Engine_load_label_from_fn(HTS_Engine * engine, char *fn);

/* HTS_Engine_load_label_from_fp: load label from file name */
void HTS_Engine_load_label_from_fp(HTS_Engine * engine, FILE * fp);

/* HTS_Engine_create_sstream: parse label and determine state duration */
void HTS_Engine_create_sstream(HTS_Engine * engine);

/* HTS_Engine_create_pstream: generate speech parameter vector sequence */
void HTS_Engine_create_pstream(HTS_Engine * engine);

/* HTS_Engine_create_gstream: synthesis speech */
void HTS_Engine_create_gstream(HTS_Engine * engine);

/* HTS_Engine_save_infomation: output trace infomation */
void HTS_Engine_save_infomation(HTS_Engine * engine, FILE * fp);

/* HTS_Engine_save_label: output label with time */
void HTS_Engine_save_label(HTS_Engine * engine, FILE * fp);

/* HTS_Engine_save_generated_parameter: output generated parameter */
void HTS_Engine_save_generated_parameter(HTS_Engine * engine, FILE * fp,
                                         int stream_index);

/* HTS_Engine_save_generated_speech: output generated speech */
void HTS_Engine_save_generated_speech(HTS_Engine * engine, FILE * fp);

/* HTS_Engine_save_riff: output RIFF format file */
void HTS_Engine_save_riff(HTS_Engine * engine, FILE * wavfp);

/* HTS_Engine_refresh: free memory per one time synthesis */
void HTS_Engine_refresh(HTS_Engine * engine);

/* HTS_Engine_clear: free engine */
void HTS_Engine_clear(HTS_Engine * engine);

/*  -------------------------- audio ------------------------------  */

#if !defined(AUDIO_PLAY_WIN32) && !defined(AUDIO_PLAY_WINCE)
#define AUDIO_PLAY_NONE
#endif                          /* !AUDIO_PLAY_WIN32 && !AUDIO_PLAY_WINCE */

/* define HTS_Audio for windows & windows mobile */
#if defined (AUDIO_PLAY_WIN32) || defined(AUDIO_PLAY_WINCE)
#include<afxwin.h>
#include<mmsystem.h>
#ifdef AUDIO_PLAY_WIN32
#pragma comment(lib,"winmm.lib")
#endif                          /* AUDIO_PLAY_WIN32 */
typedef struct _HTS_Audio {
   HWAVEOUT hwaveout;           /* audio device handle */
   WAVEFORMATEX waveformatex;   /* wave formatex */
   short *buff;                 /* current buffer */
   int buff_size;               /* current buffer size */
   int which_buff;              /* double buffering flag */
   HTS_Boolean now_buff_1;      /* double buffering flag */
   HTS_Boolean now_buff_2;      /* double buffering flag */
   WAVEHDR buff_1;              /* buffer */
   WAVEHDR buff_2;              /* buffer */
   int max_buff_size;           /* buffer size of audio output device */
} HTS_Audio;
#endif                          /* AUDIO_PLAY_WIN32 || AUDIO_PLAY_WINCE */

/* define HTS_Audio for others */
#ifdef AUDIO_PLAY_NONE
typedef struct _HTS_Audio {
   int i;                       /* make compiler happy */
} HTS_Audio;
#endif                          /* AUDIO_PLAY_NONE */

/*  -------------------------- vocoder ----------------------------  */

/* HTS_Vocoder: structure for setting of vocoder */
typedef struct _HTS_Vocoder {
   HTS_VocoderType type;
   int stage;                   /* gamma = (-1 / stage) : if stage == 0 then MLSA filter */
   double gamma;                /* gamma */
   int fprd;                    /* frame shift */
   int iprd;                    /* interpolation period */
   int seed;                    /* seed of random generator */
   unsigned long next;          /* temporary variable for random generator */
   HTS_Boolean gauss;           /* flag to use Gaussian noise */
   double rate;                 /* sampling rate */
   double p1;                   /* used in excitation generation */
   double pc;                   /* used in excitation generation */
   int sw;                      /* switch used in random generator */
   int x;                       /* excitation signal */
   HTS_Audio *audio;            /* pointer for audio device */
   double *freqt_buff;          /* used in freqt */
   int freqt_size;              /* buffer size for freqt */
   double *spectrum2en_buff;    /* used in spectrum2en */
   int spectrum2en_size;        /* buffer size for spectrum2en */
   double r1, r2, s;            /* used in random generator */
   double *postfilter_buff;     /* used in postfiltering */
   int postfilter_size;         /* buffer size for postfiltering */
   double *c, *cc, *cinc, *d1;  /* used in the MLSA/MGLSA filter */
   double *pade;                /* used in mlsadf         (MCP) */
   double *lsp2lpc_buff;        /* used in lsp2lpc        (LSP) */
   int lsp2lpc_size;            /* buffer size of lsp2lpc (LSP) */
   double *gc2gc_buff;          /* used in gc2gc          (MGC, LSP) */
   int gc2gc_size;              /* buffer size for gc2gc  (MGC, LSP) */
} HTS_Vocoder;

/*  ----------------------- vocoder method ------------------------  */

/* HTS_Vocoder_initialize: initialize vocoder */
void HTS_Vocoder_initialize(HTS_Vocoder * v, const HTS_VocoderType type,
                            const int m, const int stage, const int rate,
                            const int fperiod, int buff_size);

/* HTS_Vocoder_synthesize: pulse/noise excitation and MLSA/MGLSA filster based waveform synthesis */
void HTS_Vocoder_synthesize(HTS_Vocoder * v, const int m, double lf0,
                            double *spectrum, double alpha, double beta,
                            short *rawdata);

/* HTS_Vocoder_postfilter_mcp: postfilter for MCP */
void HTS_Vocoder_postfilter_mcp(HTS_Vocoder * v, double *mcp, const int m,
                                double alpha, double beta);

/* HTS_Vocoder_clear: clear vocoder */
void HTS_Vocoder_clear(HTS_Vocoder * v);
