/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $Id: HTS_engine.c,v 1.3 2008/02/18 04:52:53 uratec Exp $ */

/* hts_engine libraries */
#include "HTS_hidden.h"

/* HTS_Engine_initialize: initialize engine */
void HTS_Engine_initialize(HTS_Engine * engine, int nstream)
{
   int i;

   /* default value for control parameter */
   engine->global.stage = 0;
   engine->global.sampling_rate = 16000;
   engine->global.fperiod = 80;
   engine->global.rho = 0.0;
   engine->global.alpha = 0.42;
   engine->global.beta = 0.0;
   engine->global.audio_buff_size = 0;
   engine->global.type = HTS_VOCODER_TYPE_MCP;
   engine->global.msd_threshold =
       (double *) HTS_calloc(nstream, sizeof(double));
   for (i = 0; i < nstream; i++)
      engine->global.msd_threshold[i] = 0.5;

   /* interpolation weight */
   engine->global.parameter_iw =
       (double **) HTS_calloc(nstream, sizeof(double *));
   engine->global.gv_iw = (double **) HTS_calloc(nstream, sizeof(double *));
   engine->global.duration_iw = NULL;
   for (i = 0; i < nstream; i++)
      engine->global.parameter_iw[i] = NULL;
   for (i = 0; i < nstream; i++)
      engine->global.gv_iw[i] = NULL;

   /* GV weight */
   engine->global.gv_weight = (double *) HTS_calloc(nstream, sizeof(double));
   for (i = 0; i < nstream; i++)
      engine->global.gv_weight[i] = 1.0;

   /* initialize model set */
   HTS_ModelSet_initialize(&engine->ms, nstream);
   /* initialize label list */
   HTS_Label_initialize(&engine->label);
   /* initialize state sequence set */
   HTS_SStreamSet_initialize(&engine->sss);
   /* initialize pstream set */
   HTS_PStreamSet_initialize(&engine->pss);
   /* initialize gstream set */
   HTS_GStreamSet_initialize(&engine->gss);
}

/* HTS_Engine_load_duratin_from_fn: load duration pdfs, trees and number of state from file names */
void HTS_Engine_load_duration_from_fn(HTS_Engine * engine, char **pdf_fn,
                                      char **tree_fn, int interpolation_size)
{
   int i;
   FILE **pdf_fp, **tree_fp;

   pdf_fp = (FILE **) HTS_calloc(interpolation_size, sizeof(FILE *));
   tree_fp = (FILE **) HTS_calloc(interpolation_size, sizeof(FILE *));
   for (i = 0; i < interpolation_size; i++) {
      pdf_fp[i] = HTS_get_fp(pdf_fn[i], "rb");
      tree_fp[i] = HTS_get_fp(tree_fn[i], "r");
   }
   HTS_Engine_load_duration_from_fp(engine, pdf_fp, tree_fp,
                                    interpolation_size);
   for (i = 0; i < interpolation_size; i++) {
      fclose(pdf_fp[i]);
      fclose(tree_fp[i]);
   }
   HTS_free(pdf_fp);
   HTS_free(tree_fp);
}

/* HTS_Engine_load_duration_from_fp: load duration pdfs, trees and number of state from file pointers */
void HTS_Engine_load_duration_from_fp(HTS_Engine * engine, FILE ** pdf_fp,
                                      FILE ** tree_fp, int interpolation_size)
{
   int i;

   HTS_ModelSet_load_duration(&engine->ms, pdf_fp, tree_fp, interpolation_size);
   engine->global.duration_iw =
       (double *) HTS_calloc(interpolation_size, sizeof(double));
   for (i = 0; i < interpolation_size; i++)
      engine->global.duration_iw[i] = 1.0 / interpolation_size;
}

/* HTS_Engine_load_parameter_from_fn: load parameter pdfs, trees and windows from file names */
void HTS_Engine_load_parameter_from_fn(HTS_Engine * engine, char **pdf_fn,
                                       char **tree_fn, char **win_fn,
                                       int stream_index, HTS_Boolean msd_flag,
                                       int window_size, int interpolation_size)
{
   int i;
   FILE **pdf_fp, **tree_fp, **win_fp;

   pdf_fp = (FILE **) HTS_calloc(interpolation_size, sizeof(FILE *));
   tree_fp = (FILE **) HTS_calloc(interpolation_size, sizeof(FILE *));
   win_fp = (FILE **) HTS_calloc(window_size, sizeof(FILE *));
   for (i = 0; i < interpolation_size; i++) {
      pdf_fp[i] = HTS_get_fp(pdf_fn[i], "rb");
      tree_fp[i] = HTS_get_fp(tree_fn[i], "r");
   }
   for (i = 0; i < window_size; i++)
      win_fp[i] = HTS_get_fp(win_fn[i], "r");
   HTS_Engine_load_parameter_from_fp(engine, pdf_fp, tree_fp, win_fp,
                                     stream_index, msd_flag,
                                     window_size, interpolation_size);
   for (i = 0; i < interpolation_size; i++) {
      fclose(pdf_fp[i]);
      fclose(tree_fp[i]);
   }
   for (i = 0; i < window_size; i++)
      fclose(win_fp[i]);
   HTS_free(pdf_fp);
   HTS_free(tree_fp);
   HTS_free(win_fp);
}

/* HTS_Engine_load_parameter_from_fp: load parameter pdfs, trees and windows from file pointers */
void HTS_Engine_load_parameter_from_fp(HTS_Engine * engine, FILE ** pdf_fp,
                                       FILE ** tree_fp, FILE ** win_fp,
                                       int stream_index, HTS_Boolean msd_flag,
                                       int window_size, int interpolation_size)
{
   int i;

   HTS_ModelSet_load_parameter(&engine->ms, pdf_fp, tree_fp, win_fp,
                               stream_index, msd_flag,
                               window_size, interpolation_size);
   engine->global.parameter_iw[stream_index] =
       (double *) HTS_calloc(interpolation_size, sizeof(double));
   for (i = 0; i < interpolation_size; i++)
      engine->global.parameter_iw[stream_index][i] = 1.0 / interpolation_size;
}

/* HTS_Engine_load_gv_from_fn: load GV pdfs from file names */
void HTS_Engine_load_gv_from_fn(HTS_Engine * engine, char **pdf_fn,
                                int stream_index, int interpolation_size)
{
   int i;
   FILE **pdf_fp;

   pdf_fp = (FILE **) HTS_calloc(interpolation_size, sizeof(FILE *));
   for (i = 0; i < interpolation_size; i++) {
      pdf_fp[i] = HTS_get_fp(pdf_fn[i], "rb");
   }
   HTS_Engine_load_gv_from_fp(engine, pdf_fp, stream_index, interpolation_size);
   for (i = 0; i < interpolation_size; i++) {
      fclose(pdf_fp[i]);
   }
   HTS_free(pdf_fp);
}

/* HTS_Engine_load_gv_from_fp: load GV pdfs from file pointers */
void HTS_Engine_load_gv_from_fp(HTS_Engine * engine, FILE ** pdf_fp,
                                int stream_index, int interpolation_size)
{
   int i;

   HTS_ModelSet_load_gv(&engine->ms, pdf_fp, stream_index, interpolation_size);
   engine->global.gv_iw[stream_index] =
       (double *) HTS_calloc(interpolation_size, sizeof(double));
   for (i = 0; i < interpolation_size; i++)
      engine->global.gv_iw[stream_index][i] = 1.0 / interpolation_size;
}

/* HTS_Engine_set_sampling_rate: set sampling rate */
void HTS_Engine_set_sampling_rate(HTS_Engine * engine, int i)
{
   if (i < 1)
      i = 1;
   if (i > 48000)
      i = 48000;
   engine->global.sampling_rate = i;
}

/* HTS_Engine_set_fperiod: set frame shift */
void HTS_Engine_set_fperiod(HTS_Engine * engine, int i)
{
   if (i < 1)
      i = 1;
   if (i > 2000)
      i = 2000;
   engine->global.fperiod = i;
}

/* HTS_Engine_set_alpha: set alpha */
void HTS_Engine_set_alpha(HTS_Engine * engine, double f)
{
   if (f < 0.0)
      f = 0.0;
   if (f > 1.0)
      f = 1.0;
   engine->global.alpha = f;
}

/* HTS_Engine_set_gamma: set gamma (gamma = -1 / i : if i=0 then gamma=0) */
void HTS_Engine_set_gamma(HTS_Engine * engine, int i)
{
   if (i < 0)
      i = 0;
   engine->global.stage = i;
}

/* HTS_Engine_set_beta: set beta */
void HTS_Engine_set_beta(HTS_Engine * engine, double f)
{
   if (f < -0.8)
      f = -0.8;
   if (f > 0.8)
      f = 0.8;
   engine->global.beta = f;
}

/* HTS_Engine_set_rho: set control duration pparameter */
void HTS_Engine_set_rho(HTS_Engine * engine, double f)
{
   if (f < -1.0)
      f = -1.0;
   if (f > 1.0)
      f = 1.0;
   engine->global.rho = f;
}

/* HTS_Engine_set_audio_buff_size: set audio buffer size */
void HTS_Engine_set_audio_buff_size(HTS_Engine * engine, int i)
{
   if (i < 0)
      i = 0;
   if (i > 48000)
      i = 48000;
   engine->global.audio_buff_size = i;
}

/* HTS_Engine_set_vocoder_type: set vocoder type */
void HTS_Engine_set_vocoder_type(HTS_Engine * engine, HTS_VocoderType type)
{
   if (type == HTS_VOCODER_TYPE_MCP || type == HTS_VOCODER_TYPE_LSP)
      engine->global.type = type;
}

/* HTS_Egnine_set_msd_threshold: set MSD threshold */
void HTS_Engine_set_msd_threshold(HTS_Engine * engine, int stream_index,
                                  double f)
{
   if (f < 0.0)
      f = 0.0;
   if (f > 1.0)
      f = 1.0;
   engine->global.msd_threshold[stream_index] = f;
}

/* HTS_Engine_set_duration_interpolation_weight: set interpolation weight for duration */
void HTS_Engine_set_duration_interpolation_weight(HTS_Engine * engine,
                                                  int interpolation_index,
                                                  double f)
{
   engine->global.duration_iw[interpolation_index] = f;
}

/* HTS_Engine_set_parameter_interpolation_weight: set interpolation weight for parameter */
void HTS_Engine_set_parameter_interpolation_weight(HTS_Engine * engine,
                                                   int stream_index,
                                                   int interpolation_index,
                                                   double f)
{
   engine->global.parameter_iw[stream_index][interpolation_index] = f;
}

/* HTS_Engine_set_gv_interpolation_weight: set interpolation weight for GV */
void HTS_Engine_set_gv_interpolation_weight(HTS_Engine * engine,
                                            int stream_index,
                                            int interpolation_index, double f)
{
   engine->global.gv_iw[stream_index][interpolation_index] = f;
}

/* HTS_Engine_set_gv_weight: set GV weight */
void HTS_Engine_set_gv_weight(HTS_Engine * engine, int stream_index, double f)
{
   engine->global.gv_weight[stream_index] = f;
}

/* HTS_Engine_load_label_from_fn: load label from file name */
void HTS_Engine_load_label_from_fn(HTS_Engine * engine, char *fn)
{
   HTS_Label_load_from_fn(&engine->label, engine->global.sampling_rate,
                          engine->global.fperiod, fn);
}

/* HTS_Engine_load_label_from_fp: load label list from file pointer */
void HTS_Engine_load_label_from_fp(HTS_Engine * engine, FILE * fp)
{
   HTS_Label_load_from_fp(&engine->label, engine->global.sampling_rate,
                          engine->global.fperiod, fp);
}

/* HTS_Engine_create_sstream: parse label and determine state duration */
void HTS_Engine_create_sstream(HTS_Engine * engine)
{
   HTS_SStreamSet_create(&engine->sss, &engine->ms, &engine->label,
                         engine->global.rho, engine->global.duration_iw,
                         engine->global.parameter_iw, engine->global.gv_iw);
}

/* HTS_Engine_create_pstream: generate speech parameter vector sequence */
void HTS_Engine_create_pstream(HTS_Engine * engine)
{
   HTS_PStreamSet_create(&engine->pss, &engine->sss,
                         engine->global.msd_threshold,
                         engine->global.gv_weight);
}

/* HTS_Engine_create_gstream: synthesis speech */
void HTS_Engine_create_gstream(HTS_Engine * engine)
{
   HTS_GStreamSet_create(&engine->gss, &engine->pss, engine->global.type,
                         engine->global.stage, engine->global.sampling_rate,
                         engine->global.fperiod, engine->global.alpha,
                         engine->global.beta, engine->global.audio_buff_size);
}

/* HTS_Engine_save_infomation: output trace information */
void HTS_Engine_save_infomation(HTS_Engine * engine, FILE * fp)
{
   int i, j;
   double temp;
   HTS_Global *global = &engine->global;
   HTS_ModelSet *ms = &engine->ms;

   /* global parameter */
   fprintf(fp, "sampring frequency                     -> %8d(Hz)\n",
           global->sampling_rate);
   fprintf(fp, "frame period                           -> %8d(point)\n",
           global->fperiod);
   fprintf(fp, "                                          %8.5f(msec)\n",
           1e+3 * global->fperiod / global->sampling_rate);
   fprintf(fp, "all-pass constant                      -> %8.5f\n",
           (float) global->alpha);
   fprintf(fp, "gamma                                  -> %8.5f\n",
           (float) (global->stage == 0 ? 0.0 : -1.0 / global->stage));
   fprintf(fp, "postfiltering coefficient              -> %8.5f\n",
           (float) global->beta);
   fprintf(fp, "control duration parameter             -> %8.5f\n",
           (float) global->rho);
   fprintf(fp, "audio buffer size                      -> %8d(sample)\n",
           global->audio_buff_size);
   fprintf(fp, "vocoder type                           -> ");
   if (global->type == HTS_VOCODER_TYPE_MCP)
      fprintf(fp, "     MCP\n");
   else if (global->type == HTS_VOCODER_TYPE_LSP)
      fprintf(fp, "     LSP\n");
   else
      fprintf(fp, " unknown\n");

   /* duration parameter */
   fprintf(fp, "duration number of state               -> %8d\n",
           HTS_ModelSet_get_nstate(ms));
   fprintf(fp, "         interpolation size            -> %8d\n",
           HTS_ModelSet_get_duration_interpolation_size(ms));
   /* check interpolation */
   for (i = 0, temp = 0.0;
        i < HTS_ModelSet_get_duration_interpolation_size(ms); i++)
      temp += global->duration_iw[i];
   for (i = 0; i < HTS_ModelSet_get_duration_interpolation_size(ms); i++)
      global->duration_iw[i] /= temp;
   for (i = 0; i < HTS_ModelSet_get_duration_interpolation_size(ms); i++)
      fprintf(fp,
              "         interpolation weight[%2d]      -> %8.0f(%%)\n", i,
              (float) (100 * global->duration_iw[i]));

   for (i = 0; i < HTS_ModelSet_get_nstream(ms); i++) {
      /* stream parameter */
      fprintf(fp, "stream[%2d] vector length               -> %8d\n", i,
              HTS_ModelSet_get_vector_length(ms, i));
      fprintf(fp, "           dynamic window size         -> %8d\n",
              HTS_ModelSet_get_window_size(ms, i));
      /* interpolation */
      fprintf(fp, "           interpolation size          -> %8d\n",
              HTS_ModelSet_get_parameter_interpolation_size(ms, i));
      for (j = 0, temp = 0.0;
           j < HTS_ModelSet_get_parameter_interpolation_size(ms, i); j++)
         temp += global->parameter_iw[i][j];
      for (j = 0; j < HTS_ModelSet_get_parameter_interpolation_size(ms, i); j++)
         global->parameter_iw[i][j] /= temp;
      for (j = 0; j < HTS_ModelSet_get_parameter_interpolation_size(ms, i); j++)
         fprintf(fp,
                 "           interpolation weight[%2d]    -> %8.0f(%%)\n", j,
                 (float) (100 * global->parameter_iw[i][j]));
      /* MSD */
      if (HTS_ModelSet_is_msd(ms, i)) { /* for MSD */
         fprintf(fp, "           MSD flag                    ->     TRUE\n");
         fprintf(fp, "           MSD threshold               -> %8.5f\n",
                 global->msd_threshold[i]);
      } else {                  /* for non MSD */
         fprintf(fp, "           MSD flag                    ->    FALSE\n");
      }
      /* GV */
      if (HTS_ModelSet_use_gv(ms, i)) {
         fprintf(fp, "           GV flag                     ->     TRUE\n");
         fprintf(fp, "           GV weight                   -> %8.0f(%%)\n",
                 (float) (100 * global->gv_weight[i]));
         fprintf(fp, "           GV interpolation size       -> %8d\n",
                 HTS_ModelSet_get_gv_interpolation_size(ms, i));
         /* interpolation */
         for (j = 0, temp = 0.0;
              j < HTS_ModelSet_get_gv_interpolation_size(ms, i); j++)
            temp += global->gv_iw[i][j];
         for (j = 0; j < HTS_ModelSet_get_gv_interpolation_size(ms, i); j++)
            global->gv_iw[i][j] /= temp;
         for (j = 0; j < HTS_ModelSet_get_gv_interpolation_size(ms, i); j++)
            fprintf(fp,
                    "           GV interpolation weight[%2d] -> %8.0f(%%)\n", j,
                    (float) (100 * global->gv_iw[i][j]));
      } else {
         fprintf(fp, "           GV flag                     ->    FALSE\n");
      }
   }
}

/* HTS_Engine_save_label: output label with time */
void HTS_Engine_save_label(HTS_Engine * engine, FILE * fp)
{
   int i, j;
   int frame, state, duration;

   HTS_Label *label = &engine->label;
   HTS_SStreamSet *sss = &engine->sss;
   const int nstate = HTS_ModelSet_get_nstate(&engine->ms);
   const double rate =
       engine->global.fperiod * 1e+7 / engine->global.sampling_rate;

   for (i = 0, state = 0, frame = 0; i < HTS_Label_get_size(label); i++) {
      for (j = 0, duration = 0; j < nstate; j++)
         duration += HTS_SStreamSet_get_duration(sss, state++);
      /* in HTK & HTS format */
      fprintf(fp, "%d %d %s\n", (int) (frame * rate),
              (int) ((frame + duration) * rate),
              HTS_Label_get_string(label, i));
      frame += duration;
   }
}

/* HTS_Engine_save_generated_parameter: output generated parameter */
void HTS_Engine_save_generated_parameter(HTS_Engine * engine, FILE * fp,
                                         int stream_index)
{
   int i, j;
   float temp;
   HTS_GStreamSet *gss = &engine->gss;

   for (i = 0; i < HTS_GStreamSet_get_total_frame(gss); i++)
      for (j = 0; j < HTS_GStreamSet_get_static_length(gss, stream_index); j++) {
         temp = (float) HTS_GStreamSet_get_parameter(gss, stream_index, i, j);
         fwrite(&temp, sizeof(float), 1, fp);
      }
}

/* HTS_Engine_save_generated_speech: output generated speech */
void HTS_Engine_save_generated_speech(HTS_Engine * engine, FILE * fp)
{
   int i;
   short temp;
   HTS_GStreamSet *gss = &engine->gss;

   for (i = 0; i < HTS_GStreamSet_get_total_nsample(gss); i++) {
      temp = HTS_GStreamSet_get_speech(gss, i);
      fwrite(&temp, sizeof(short), 1, fp);
   }
}

/* HTS_Engine_save_riff: output RIFF format file */
void HTS_Engine_save_riff(HTS_Engine * engine, FILE * fp)
{
   int i;
   short temp;

   HTS_GStreamSet *gss = &engine->gss;
   char data_01_04[] = { 'R', 'I', 'F', 'F' };
   int data_05_08 = HTS_GStreamSet_get_total_nsample(gss) * sizeof(short) + 36;
   char data_09_12[] = { 'W', 'A', 'V', 'E' };
   char data_13_16[] = { 'f', 'm', 't', ' ' };
   int data_17_20 = 16;
   short data_21_22 = 1;        /* PCM */
   short data_23_24 = 1;        /* monoral */
   int data_25_28 = engine->global.sampling_rate;
   int data_29_32 = engine->global.sampling_rate * sizeof(short);
   short data_33_34 = sizeof(short);
   short data_35_36 = (short) (sizeof(short) * 8);
   char data_37_40[] = { 'd', 'a', 't', 'a' };
   int data_41_44 = HTS_GStreamSet_get_total_nsample(gss) * sizeof(short);

   /* write header */
   HTS_fwrite_little_endian(data_01_04, sizeof(char), 4, fp);
   HTS_fwrite_little_endian(&data_05_08, sizeof(int), 1, fp);
   HTS_fwrite_little_endian(data_09_12, sizeof(char), 4, fp);
   HTS_fwrite_little_endian(data_13_16, sizeof(char), 4, fp);
   HTS_fwrite_little_endian(&data_17_20, sizeof(int), 1, fp);
   HTS_fwrite_little_endian(&data_21_22, sizeof(short), 1, fp);
   HTS_fwrite_little_endian(&data_23_24, sizeof(short), 1, fp);
   HTS_fwrite_little_endian(&data_25_28, sizeof(int), 1, fp);
   HTS_fwrite_little_endian(&data_29_32, sizeof(int), 1, fp);
   HTS_fwrite_little_endian(&data_33_34, sizeof(short), 1, fp);
   HTS_fwrite_little_endian(&data_35_36, sizeof(short), 1, fp);
   HTS_fwrite_little_endian(data_37_40, sizeof(char), 4, fp);
   HTS_fwrite_little_endian(&data_41_44, sizeof(int), 1, fp);
   /* write data */
   for (i = 0; i < HTS_GStreamSet_get_total_nsample(gss); i++) {
      temp = HTS_GStreamSet_get_speech(gss, i);
      HTS_fwrite_little_endian(&temp, sizeof(short), 1, fp);
   }
}

/* HTS_Engine_refresh: free model per one time synthesis */
void HTS_Engine_refresh(HTS_Engine * engine)
{
   /* free generated parameter stream set */
   HTS_GStreamSet_clear(&engine->gss);
   /* free parameter stream set */
   HTS_PStreamSet_clear(&engine->pss);
   /* free state stream set */
   HTS_SStreamSet_clear(&engine->sss);
   /* free label list */
   HTS_Label_clear(&engine->label);
}

/* HTS_Engine_clear: free engine */
void HTS_Engine_clear(HTS_Engine * engine)
{
   int i;

   HTS_free(engine->global.msd_threshold);
   HTS_free(engine->global.duration_iw);
   for (i = 0; i < HTS_ModelSet_get_nstream(&engine->ms); i++) {
      HTS_free(engine->global.parameter_iw[i]);
      if (engine->global.gv_iw[i])
         HTS_free(engine->global.gv_iw[i]);
   }
   HTS_free(engine->global.parameter_iw);
   HTS_free(engine->global.gv_iw);
   HTS_free(engine->global.gv_weight);

   HTS_ModelSet_clear(&engine->ms);
}
