/*  ---------------------------------------------------------------  */
/*           The HMM-Based Speech Synthesis System (HTS)             */
/*                       HTS Working Group                           */
/*                                                                   */
/*                  Department of Computer Science                   */
/*                  Nagoya Institute of Technology                   */
/*                               and                                 */
/*   Interdisciplinary Graduate School of Science and Engineering    */
/*                  Tokyo Institute of Technology                    */
/*                                                                   */
/*                     Copyright (c) 2001-2007                       */
/*                       All Rights Reserved.                        */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The source code must retain the above copyright notice,     */
/*       this list of conditions and the following disclaimer.       */
/*                                                                   */
/*    2. Any modifications to the source code must be clearly        */
/*       marked as such.                                             */
/*                                                                   */
/*    3. Redistributions in binary form must reproduce the above     */
/*       copyright notice, this list of conditions and the           */
/*       following disclaimer in the documentation and/or other      */
/*       materials provided with the distribution.  Otherwise, one   */
/*       must contact the HTS working group.                         */
/*                                                                   */
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF TECHNOLOGY,   */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSTITUTE OF         */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */

/* $id: HTS_label.c,v 1.1.2.1 2008/02/12 10:47:43 uratec Exp $ */

#include <stdlib.h>             /* for atoi() */
#include <ctype.h>              /* for isgraph(),isdigit() */

/* hts_engine libraries */
#include "HTS_hidden.h"

/*
  [static function]

  [non static function]

  HTS_Label_initialize
  HTS_Label_load_from_fn
  HTS_Label_load_from_fp
  HTS_Label_load_from_string
  HTS_Label_load_from_string_list
  HTS_Label_set_speech_speed
  HTS_Label_set_frame
  HTS_Label_get_size
  HTS_Label_get_string
  HTS_Label_clear
*/

/* HTS_Label_initialize: initialize label */
void HTS_Label_initialize(HTS_Label * label)
{
   label->head = NULL;
   label->size = 0;
}

/* HTS_Label_load_from_fn: load label from file name */
void HTS_Label_load_from_fn(HTS_Label * label, int sampling_rate, int fperiod,
                            char *fn)
{
   FILE *fp = HTS_get_fp(fn, "r");
   HTS_Label_load_from_fp(label, sampling_rate, fperiod, fp);
   fclose(fp);
}

/* HTS_Label_load_from_fp: load label from file pointer */
void HTS_Label_load_from_fp(HTS_Label * label, int sampling_rate, int fperiod,
                            FILE * fp)
{
   char buff[HTS_MAXBUFLEN];
   HTS_LabelString *lstring = NULL;
   int start, end;
   const double rate = sampling_rate / (fperiod * 1e+7);

   if (label->head || label->size != 0)
      HTS_error(1, "HTS_Label_load_from_fp: label has to be initialized.");
   /* parse label file */
   while (HTS_get_token(fp, buff)) {
      if (!isgraph(buff[0]))
         break;
      label->size++;

      if (lstring) {
         lstring->next =
             (HTS_LabelString *) HTS_calloc(1, sizeof(HTS_LabelString));
         lstring = lstring->next;
      } else {                  /* first time */
         lstring = (HTS_LabelString *) HTS_calloc(1, sizeof(HTS_LabelString));
         label->head = lstring;
      }
      if (isdigit(buff[0])) {   /* has frame infomation */
         start = atoi(buff);
         HTS_get_token(fp, buff);
         end = atoi(buff);
         HTS_get_token(fp, buff);
         lstring->frame_flag = TRUE;
         lstring->frame = (int) (rate * (end - start));
         if (lstring->frame < 1)
            lstring->frame = 1;
      } else {
         lstring->frame_flag = FALSE;
         lstring->frame = 0;
      }
      lstring->next = NULL;
      lstring->name = HTS_strdup(buff);
      lstring->speed_flag = FALSE;
      lstring->speed = 0.0;
   }
}

/* HTS_Label_load_from_string: load label from string */
void HTS_Label_load_label_from_string(HTS_Label * label, int sampling_rate,
                                      int fperiod, char *data)
{
   char buff[HTS_MAXBUFLEN];
   HTS_LabelString *lstring = NULL;
   int data_index = 0;          /* data index */
   int start, end;
   const double rate = sampling_rate / (fperiod * 1e+7);

   if (label->head || label->size != 0)
      HTS_error(1, "HTS_Label_load_from_fp: label list has to be initialized.");
   /* copy label */
   while (HTS_get_token_from_string(data, &data_index, buff)) {
      if (!isgraph(buff[0]))
         break;
      label->size++;

      if (lstring) {
         lstring->next =
             (HTS_LabelString *) HTS_calloc(1, sizeof(HTS_LabelString));
         lstring = lstring->next;
      } else {                  /* first time */
         lstring = (HTS_LabelString *) HTS_calloc(1, sizeof(HTS_LabelString));
         label->head = lstring;
      }
      if (isdigit(buff[0])) {   /* has frame infomation */
         start = atoi(buff);
         HTS_get_token_from_string(data, &data_index, buff);
         end = atoi(buff);
         HTS_get_token_from_string(data, &data_index, buff);
         lstring->frame_flag = TRUE;
         lstring->frame = (int) (rate * (end - start));
         if (lstring->frame < 1)
            lstring->frame = 1;
      } else {
         lstring->frame_flag = FALSE;
         lstring->frame = 0;
      }
      lstring->next = NULL;
      lstring->name = HTS_strdup(buff);
      lstring->speed_flag = FALSE;
      lstring->speed = 0.0;
   }
}

/* HTS_Label_load_from_string_list: load label from string list */
void HTS_Label_load_from_string_list(HTS_Label * label, int sampling_rate,
                                     int fperiod, char **data, int size)
{
   char buff[HTS_MAXBUFLEN];
   HTS_LabelString *lstring = NULL;
   int i;
   int data_index;
   int start, end;
   const double rate = sampling_rate / (fperiod * 1e+7);

   if (label->head || label->size != 0)
      HTS_error(1, "HTS_Label_load_from_fp: label list has to be initialized.");
   /* copy label */
   for (i = 0; i < size; i++) {
      if (!isgraph(data[i][0]))
         break;
      label->size++;

      if (lstring) {
         lstring->next =
             (HTS_LabelString *) HTS_calloc(1, sizeof(HTS_LabelString));
         lstring = lstring->next;
      } else {                  /* first time */
         lstring = (HTS_LabelString *) HTS_calloc(1, sizeof(HTS_LabelString));
         label->head = lstring;
      }
      data_index = 0;
      if (isdigit(data[i][0])) {        /* has frame infomation */
         HTS_get_token_from_string(data[i], &data_index, buff);
         start = atoi(buff);
         HTS_get_token_from_string(data[i], &data_index, buff);
         end = atoi(buff);
         HTS_get_token_from_string(data[i], &data_index, buff);
         lstring->frame_flag = TRUE;
         lstring->frame = (int) (rate * (end - start));
         if (lstring->frame < 1)
            lstring->frame = 1;
      } else {
         lstring->frame_flag = FALSE;
         lstring->frame = 0;
      }
      lstring->next = NULL;
      lstring->name = HTS_strdup(&buff[data_index]);
      lstring->speed_flag = FALSE;
      lstring->speed = 0.0;
   }
}

/* HTS_Label_set_speech_speed: set speech speed */
void HTS_Label_set_speech_speed(HTS_Label * label, int string_index, double f)
{
   HTS_LabelString *lstring = label->head;

   while (string_index-- && lstring)
      lstring = lstring->next;
   if (!lstring)
      return;
   if (f >= 0.2 && f <= 1.8) {
      lstring->speed_flag = TRUE;
      lstring->speed = f;
   } else {
      lstring->speed_flag = FALSE;
      lstring->speed = 0.0;
   }
}

/* HTS_Label_set_frame: set frame length */
void HTS_Label_set_frame(HTS_Label * label, int string_index, int i)
{
   HTS_LabelString *lstring = label->head;

   while (string_index-- && lstring)
      lstring = lstring->next;
   if (!lstring)
      return;
   if (i >= 1) {
      lstring->frame_flag = TRUE;
      lstring->frame = i;
   } else {
      lstring->frame_flag = FALSE;
      lstring->frame = 0;
   }
}

/* HTS_Label_get_size: get number of label string */
inline int HTS_Label_get_size(HTS_Label * label)
{
   return label->size;
}

/* HTS_Label_get_string: get label string */
char *HTS_Label_get_string(HTS_Label * label, int string_index)
{
   HTS_LabelString *lstring = label->head;

   while (string_index-- && lstring)
      lstring = lstring->next;
   if (!lstring)
      return NULL;
   return lstring->name;
}

/* HTS_Label_get_frame_specified_flag: get frame specified flag */
HTS_Boolean HTS_Label_get_frame_specified_flag(HTS_Label * label,
                                               int string_index)
{
   HTS_LabelString *lstring = label->head;

   while (string_index-- && lstring)
      lstring = lstring->next;
   if (!lstring)
      return FALSE;
   return lstring->frame_flag;
}

/* HTS_Label_get_frame: get frame length*/
int HTS_Label_get_frame(HTS_Label * label, int string_index)
{
   HTS_LabelString *lstring = label->head;

   while (string_index-- && lstring)
      lstring = lstring->next;
   if (!lstring)
      return 0;
   return lstring->frame;
}

/* HTS_Label_get_speed_specified_flag: get speech speed specified flag */
HTS_Boolean HTS_Label_get_speed_specified_flag(HTS_Label * label,
                                               int string_index)
{
   HTS_LabelString *lstring = label->head;

   while (string_index-- && lstring)
      lstring = lstring->next;
   if (!lstring)
      return FALSE;
   return lstring->speed_flag;
}

/* HTS_Label_get_speed: get frame length*/
double HTS_Label_get_speed(HTS_Label * label, int string_index)
{
   HTS_LabelString *lstring = label->head;

   while (string_index-- && lstring)
      lstring = lstring->next;
   if (!lstring)
      return 1.0;
   return lstring->speed;
}

/* HTS_Label_clear: free label */
void HTS_Label_clear(HTS_Label * label)
{
   HTS_LabelString *lstring, *next_lstring;

   for (lstring = label->head; lstring; lstring = next_lstring) {
      next_lstring = lstring->next;
      HTS_free(lstring->name);
      HTS_free(lstring);
   }
   HTS_Label_initialize(label);
}
